#include <stdio.h>
#include <stdlib.h>
#include "align.h"
#include "heap.h"

#ifndef UNIT_HEAP_SIZE
#ifdef SAVE_MEMORY
#define UNIT_HEAP_SIZE 100000L
#else
#define UNIT_HEAP_SIZE 500000L
#endif
#endif

#define parent(n) (((n) + 1) / 2 - 1)
#define left_child(n) (((n) + 1) * 2 - 1)
#define right_child(n) (((n) + 1) * 2)

pqueue_t pqueue_init(int dir)
{
  heap_t heap;

  if((heap = (heap_t)malloc(sizeof(struct heap))) == NULL){
    fprintf(stderr, "Can't allocate memory for heap\n");
    exit(1);
  }
  heap->dir = dir;
  heap->heap_size = UNIT_HEAP_SIZE;
  heap->node_count = 0;
#ifdef DEBUG
  heap->max_node_count = 0;
#endif
  if((heap->node_table
      = (node_t *)malloc(sizeof(node_t) * (heap->heap_size))) == NULL){
    fprintf(stderr, "Can't allocate memory for heap\n");
    exit(1);
  }
  return (pqueue_t)heap;
}

void pqueue_finish(pqueue_t pqueue)
{
#ifdef DEBUG
  fprintf(stderr, "Final priority queue size = %d\n",
	  ((heap_t)pqueue)->node_count);
  fprintf(stderr, "Max priority queue size = %d\n",
	  ((heap_t)pqueue)->max_node_count);
#endif
  free(((heap_t)pqueue)->node_table);
  free((heap_t)pqueue);
}

void reconstruct_heap_downward(heap_t heap, int n)
{
  int i, j;
  node_t t;

  for(i = n; left_child(i) < heap->node_count; i = j){
    j = right_child(i);
    if((j > heap->node_count) ||
       (lookup_cost(heap->node_table[left_child(i)], heap->dir)
	< lookup_cost(heap->node_table[j], heap->dir))){
      j = left_child(i);
    }
    if(lookup_cost(heap->node_table[j], heap->dir)
       >= lookup_cost(heap->node_table[i], heap->dir)){
      break;
    }
    set_pqueue_index(heap->node_table[i], heap->dir, j);
    set_pqueue_index(heap->node_table[j], heap->dir, i);
    t = heap->node_table[j];
    heap->node_table[j] = heap->node_table[i];
    heap->node_table[i] = t;
  }
}

void reconstruct_heap_upward(heap_t heap, int n)
{
  int i;
  node_t t;

  for(i = n;
      (i > 0) && (lookup_cost(heap->node_table[i], heap->dir)
		  < lookup_cost(heap->node_table[parent(i)], heap->dir));
      i = parent(i)){
    set_pqueue_index(heap->node_table[i], heap->dir, parent(i));
    set_pqueue_index(heap->node_table[parent(i)], heap->dir, i);
    t = heap->node_table[parent(i)];
    heap->node_table[parent(i)] = heap->node_table[i];
    heap->node_table[i] = t;
  }
}

void put_node_to_pqueue(pqueue_t pqueue, node_t node)
{
  if(++((heap_t)pqueue)->node_count == ((heap_t)pqueue)->heap_size){
    ((heap_t)pqueue)->heap_size += UNIT_HEAP_SIZE;
    if((((heap_t)pqueue)->node_table
	= (node_t *)realloc(((heap_t)pqueue)->node_table,
			    sizeof(node_t) * (((heap_t)pqueue)->heap_size)))
       == NULL){
      fprintf(stderr, "Can't allocate memory for heap\n");
      exit(1);
    }
#ifdef VERBOSE
    fprintf(stderr, "Re-allocated memory for heap\n");
#endif
  }
#ifdef DEBUG
  if(((heap_t)pqueue)->max_node_count < ((heap_t)pqueue)->node_count){
    ((heap_t)pqueue)->max_node_count = ((heap_t)pqueue)->node_count;
  }
#endif
  set_pqueue_index(node, ((heap_t)pqueue)->dir,
		   ((heap_t)pqueue)->node_count - 1);
  ((heap_t)pqueue)->node_table[((heap_t)pqueue)->node_count - 1] = node;
  reconstruct_heap_upward((heap_t)pqueue, ((heap_t)pqueue)->node_count - 1);
}

node_t get_node_from_pqueue(pqueue_t pqueue)
{
  node_t t;

  if(--((heap_t)pqueue)->node_count < 0){
    return NULL;
  }
  t = ((heap_t)pqueue)->node_table[0];
  set_pqueue_index(((heap_t)pqueue)->node_table[((heap_t)pqueue)->node_count],
		   ((heap_t)pqueue)->dir, 0);
  ((heap_t)pqueue)->node_table[0]
    = ((heap_t)pqueue)->node_table[((heap_t)pqueue)->node_count];
  reconstruct_heap_downward((heap_t)pqueue, 0);
  return t;
}

void update_pqueue(pqueue_t pqueue, node_t node)
{
  reconstruct_heap_upward((heap_t)pqueue,
			  lookup_pqueue_index(node, ((heap_t)pqueue)->dir));
}

void print_pqueue(pqueue_t pqueue)
{
  int i, j;

  i = 0;
  for(j = 0; j < ((heap_t)pqueue)->node_count; j++){
    printf("%d ", lookup_cost(((heap_t)pqueue)->node_table[j],
			      ((heap_t)pqueue)->dir));
    if(j == i){
      printf("\n");
      i = left_child(i);
    }
  }
  printf("\n");
}

void extend_node(pqueue_t pqueue, node_t node, node_t next_node, int edge_cost)
{ 
  if(lookup_status(next_node, ((heap_t)pqueue)->dir) != FINISHED){
    if(lookup_status(next_node, ((heap_t)pqueue)->dir) == VISITED){
      if(lookup_cost(node, ((heap_t)pqueue)->dir) + edge_cost
	 < lookup_cost(next_node, ((heap_t)pqueue)->dir)){
	set_cost(next_node, ((heap_t)pqueue)->dir,
		 lookup_cost(node, ((heap_t)pqueue)->dir) + edge_cost);
	set_prev_node(next_node, ((heap_t)pqueue)->dir, node);
	update_pqueue(pqueue, next_node);
      }
    }else{
      set_status(next_node, ((heap_t)pqueue)->dir, VISITED);
      set_cost(next_node, ((heap_t)pqueue)->dir,
	       lookup_cost(node, ((heap_t)pqueue)->dir) + edge_cost);
      set_prev_node(next_node, ((heap_t)pqueue)->dir, node);
      put_node_to_pqueue(pqueue, next_node);
    }
  }else{
    if(lookup_cost(node, ((heap_t)pqueue)->dir) + edge_cost
       < lookup_cost(next_node, ((heap_t)pqueue)->dir)){
      set_status(next_node, ((heap_t)pqueue)->dir, VISITED);
      set_cost(next_node, ((heap_t)pqueue)->dir,
	       lookup_cost(node, ((heap_t)pqueue)->dir) + edge_cost);
      set_prev_node(next_node, ((heap_t)pqueue)->dir, node);
      put_node_to_pqueue(pqueue, next_node);
    }
  }    
}
